"""
Contains code for interacting with the MusicBrainz API.
"""

import asyncio
import os
import traceback
import urllib
import urllib.parse
import urllib.request
from typing import List, Self, Optional, Tuple

from gi.repository import GObject

from ._async import event_loop
from .utils.queuedl import EartagQueuedDownloader, EartagDownloaderMode
from .utils.misc import simplify_compare
from .backends.file import EartagFile, CoverType
from .logger import logger

try:
    # HACK: The Gst backend of audioread, which is used by acoustid, is not very
    # happy when the GLib async event loop policy is set up.
    # Force-disable the Gst backend.
    import audioread

    audioread._gst_available = lambda: False
except ImportError:
    fpcalc = os.environ.get("FPCALC", "fpcalc")
    import shutil

    if not shutil.which(fpcalc):
        logger.warning(
            "Neither audioread nor fpcalc are available, acoustid matches may be affected"
        )

import acoustid

try:
    from . import ACOUSTID_API_KEY
except ImportError:  # handle test suite import
    from tests.common import ACOUSTID_API_KEY, config
else:
    from .config import config

from .utils import simplify_string, title_case_preserve_uppercase

LAST_REQUEST = 0


mb_query = EartagQueuedDownloader(
    EartagDownloaderMode.MODE_JSON, simultaneous_downloads=1, throttle=True
)


def build_url(endpoint, id="", **kwargs):
    """Builds a MusicBrainz API endpoint URL."""
    args = []
    for argname, argdata in kwargs.items():
        if isinstance(argdata, list) or isinstance(argdata, tuple):
            argdata = "+".join(argdata)
        elif argname == "query" and isinstance(argdata, dict):
            argdata = " AND ".join([f"{k}:{v}" for k, v in argdata.items() if v])
        args.append(f"{argname}={urllib.parse.quote(argdata, encoding='utf-8')}")
    if id:
        return f"https://musicbrainz.org/ws/2/{endpoint}/{id}?{'&'.join(args)}&fmt=json"
    return f"https://musicbrainz.org/ws/2/{endpoint}?{'&'.join(args)}&fmt=json"


class EartagCAACover(GObject.Object):
    """Represents a cover art from the Cover Art Archive."""

    path = GObject.Property(type=str, default="")
    loaded = GObject.Property(type=bool, default=False)

    cover_downloader = EartagQueuedDownloader(mode=EartagDownloaderMode.MODE_FILE)

    def __init__(self, item_type: str, item_id: str, cover_type: str, cover_size: int = -1):
        """
        Initialize the CAACover object.

        :param item_type: Type of item to download the cover for (release, release-group)
        :param item_id: ID of the item to download the cover for
        :param cover_type: Type of the cover to download: thumbnail, front, back
        :param cover_size: Size of the cover to download: 0, 250, 500, 1200, 2000 (max size)
        """
        super().__init__()

        if cover_size == -1:
            cover_size = int(config.get_enum("musicbrainz-cover-size"))
        if cover_size not in (0, 250, 500, 1200, 2000):
            raise ValueError("Cover size must be one of 0, 250, 500, 1200, 2000")
        if cover_type not in ("thumbnail", "front", "back"):
            raise ValueError("Cover type must be one of thumbnail, front, back")
        if cover_type == "back" and item_type == "release-group":
            raise ValueError("Back cover is not supported for release groups")

        self.item_type = item_type
        self.item_id = item_id

        if cover_type == "thumbnail":
            self.cover_type = "front"
            self.cover_size = 250
        else:
            self.cover_type = cover_type
            self.cover_size = cover_size

    @property
    def url(self):
        """The URL to the cover."""
        url = f"https://coverartarchive.org/{self.item_type}/{self.item_id}/{self.cover_type}"
        if self.cover_size not in (0, 2000):
            url += f"-{self.cover_size}"
        return url

    async def download(self):
        try:
            tempfile = await EartagCAACover.cover_downloader.download(self.url)
        except Exception as e:
            print(e)
            tempfile = False
        if tempfile:
            self.props.path = tempfile.name
        self.props.loaded = True

    @classmethod
    def clear_tempfiles(cls):
        return cls.cover_downloader.clear_tempfiles()


class MusicBrainzRecording(GObject.Object):
    __gtype_name__ = "MusicBrainzRecording"

    SELECT_RELEASE_FIRST = -1

    def __init__(self, data: dict):
        super().__init__()
        self._recording_id = data["id"]
        self.mb_data = data
        self._available_releases = None
        self._release = MusicBrainzRecording.SELECT_RELEASE_FIRST

    def dispose(self):
        self.release.dispose()

    @property
    def available_releases(self):
        if self._available_releases is None:
            self._fetch_available_releases()
        return self._available_releases

    def _fetch_available_releases(self):
        """Fill the self.available_releases list with releases (as MusicBrainzRelease objects)."""
        if "releases" not in self.mb_data:
            logger.error("No releases in data, this should never happen!")
            return []

        # HACK: Sometimes releases don't have the "artist-credit"; in those cases,
        # try to get the artist credit from the release group, and if there is none,
        # fill both it and the release credit with the recording credit.
        for rel in self.mb_data["releases"]:
            if "artist-credit" not in rel:
                if "artist-credit" not in rel["release-group"]:
                    rel["release-group"]["artist-credit"] = self.mb_data["artist-credit"]
                rel["artist-credit"] = rel["release-group"]["artist-credit"]

        available_releases = [MusicBrainzRelease(rel) for rel in self.mb_data["releases"]]

        self._available_releases = available_releases

    @staticmethod
    def sort_recordings(
        recordings: list["MusicBrainzRecording"],
        file: Optional[EartagFile] = None,
        overrides: Optional[dict[str, str]] = None,
    ):
        """Sort a list of recordings by usefulness."""

        ret = recordings.copy()

        def dict_diff(dict_a, dict_b, show_value_diff=True):
            result = {}
            result["added"] = {k: dict_b[k] for k in set(dict_b) - set(dict_a)}
            result["removed"] = {k: dict_a[k] for k in set(dict_a) - set(dict_b)}
            if show_value_diff:
                common_keys = set(dict_a) & set(dict_b)
                result["value_diffs"] = {
                    k: (dict_a[k], dict_b[k]) for k in common_keys if dict_a[k] != dict_b[k]
                }
            return result

        # Move video recordings to the end
        ret.sort(key=lambda rec: int(rec.mb_data.get("video", False) or False))

        # Prefer recordings with earlier release date; tracks with no release date are moved to the end
        ret.sort(key=lambda rec: rec.releasedate or "Z")

        # Prefer recordings with more legitimate releases
        def _release_sort(rec) -> int:
            out = 0
            for rel in rec.mb_data.get("releases", []):
                if rel.get("status", "Official") == "Official" and "Compilation" not in rel.get(
                    "release_group", {}
                ).get("secondary-types", []):
                    out += 1
            return out

        ret.sort(key=lambda rec: _release_sort(rec), reverse=True)

        # Sort the resulting recordings by which one matches our file the best
        if file or overrides:

            def _rec_file_cmp(rec) -> int:
                out = 0
                for prop in (
                    "artist",
                    "album",
                    "releasedate",
                    "tracknumber",
                    "totaltracknumber",
                ):
                    if overrides and prop in overrides:
                        value = overrides[prop]
                    elif file and file.has_tag(prop):
                        value = file.get_property(prop)
                    else:
                        continue

                    rec_value = rec.get_property(prop)

                    if prop in ("artist", "album"):
                        out += (int(value == rec_value) * 2) + int(
                            simplify_compare(value, rec_value)
                        )
                    else:
                        out += int(value == rec_value)

                # Titles are handled separately, since a single recording may have
                # multiple alternative titles in various releases
                title = None
                if overrides:
                    title = overrides.get("title", None)
                if not title and file.has_tag("title"):
                    title = file.title

                if title:
                    out += int(title in rec.all_titles) * 2
                    out += int(
                        simplify_string(title) in [simplify_string(s) for s in rec.all_titles]
                    )

                return out

            ret.sort(key=_rec_file_cmp, reverse=True)  # The larger the number, the more tag matches

        return ret

    @staticmethod
    async def new_for_id(recording_id: str) -> Self:
        """Create a new MusicBrainzRecording object with the given ID."""
        data = await mb_query.download(
            build_url(
                "recording",
                recording_id,
                inc=("releases", "genres", "artist-credits", "media", "release-groups"),
            )
        )
        return MusicBrainzRecording(data)

    @staticmethod
    async def get_recordings_for_file(
        file, overrides: dict = None, sort: bool = True
    ) -> List[Self]:
        """
        Search for recordings matching information extracted from the file.

        Returns a list of MusicBrainzRecording files.
        """

        if overrides:
            title = overrides.get("title", file.props.title)
            artist = overrides.get("artist", file.props.artist)
            album = overrides.get("album", file.props.album)
        else:
            title = file.props.title
            artist = file.props.artist
            album = file.props.album

        if not title:
            raise ValueError("Not enough data for query; title is needed")

        logger.debug(
            f"Getting recordings for file {os.path.basename(file.path)}; overrides: {overrides}"
        )

        # Query MusicBrainz based on the available metadata
        async def _query_recordings(_title, _artist, _album):
            _data = await mb_query.download(
                build_url(
                    "recording",
                    "",
                    query={
                        "recording": _title,
                        "artist": _artist,
                        "release": _album,
                    },
                )
            )
            return [
                r
                for r in _data.get("recordings", [])
                if r.get("score") >= config["musicbrainz-confidence-treshold"]
                or simplify_string(r.get("title", "")) == simplify_string(_title)
                and simplify_string(r.get("artist-credit", [{}])[0].get("name", ""))
                == simplify_string(_artist)
            ]

        search_data = []
        search_data_ids = set()
        # For convenience, each possible "method" of querying is numbered;
        # the number increases with each query iteration. This is done so that
        # once one of the methods works, we can quickly break out of the loop.
        fetch_method = 0
        got_useful_result = False
        while True:
            # Method 0. Perform a regular query with all the parameters.
            if fetch_method == 0:
                logger.debug("  - Method 0: regular query with all parameters")
                new_search_data = await _query_recordings(title, artist, album)

            # Method 1. Perform a query without the album, if we are given one.
            elif fetch_method == 1:
                if album:
                    logger.debug("  - Method 1: query with no album")
                    new_search_data = await _query_recordings(title, artist, "")
                else:
                    logger.debug("  - Method 1 (skipped): query with no album")

            # Method 2. Simplify title and artist.
            elif fetch_method == 2:
                if (
                    simplify_string(title) != title
                    and simplify_string(artist) != artist
                    and (simplify_string(album) != album or not album)
                ):
                    logger.debug("  - Method 2: query with simplified tags")
                    new_search_data = await _query_recordings(
                        simplify_string(title),
                        simplify_string(artist),
                        simplify_string(album),
                    )
                else:
                    logger.debug("  - Method 2 (skipped): query with simplified tags")

            # Method 3. Same as 2, but without album, if we are given one.
            elif fetch_method == 3:
                if album and simplify_string(title) != title and simplify_string(artist) != artist:
                    logger.debug("  - Method 3: query with simplified tags and no album")
                    new_search_data = await _query_recordings(
                        simplify_string(title), simplify_string(artist), ""
                    )
                else:
                    logger.debug("  - Method 3 (skipped): query with simplified tags and no album")

            # Once we have exhausted all methods, return empty data.
            else:
                logger.debug("  - Ran out of methods.")
                break

            search_data += [r for r in new_search_data if r.get("id") not in search_data_ids]
            search_data_ids = search_data_ids.union(set([r.get("id") for r in new_search_data]))
            fetch_method += 1

            logger.debug(f"    Found {len(new_search_data)} new results (total {len(search_data)})")

            # Filter out non-useful results; if we didn't get anything that seems
            # correct, continue to the next method.
            for r in new_search_data:
                for _rel_data in r.get("releases", []):
                    if _rel_data.get("status", "Official") == "Official":
                        if "Compilation" not in _rel_data.get("release-group", {}).get(
                            "secondary-types", []
                        ):
                            got_useful_result = True
                            break

            if got_useful_result:
                logger.debug("  ... got useful result!")
                break
            else:
                logger.debug("  ... no useful result, trying next method")

        # Convert the search results to MusicBrainzRecording objects
        ret = []
        for r in search_data:
            try:
                rec = MusicBrainzRecording(r)
            except:  # noqa: E722
                logger.error("Error while parsing MusicBrainz recording data")
                traceback.print_exc()
                continue

            # Sort releases in the recording
            rec.sort_releases(file=file)

            ret.append(rec)

        logger.debug(f"Found {len(ret)} recordings")
        logger.debug(f"Recordings before sorting: \n{ret}")

        if sort:
            # Sort the recordings by usefulness.
            ret = MusicBrainzRecording.sort_recordings(ret, file=file, overrides=overrides)

        logger.debug(f"Recordings after sorting: \n{ret}")

        return ret

    def sort_releases(self, file: Optional[EartagFile] = None):
        """Sort available releases and pick the best matching one."""
        album = file.props.album if file else ""
        releasedate = file.props.releasedate if file else ""
        totaltracknumber = file.props.totaltracknumber if file else ""

        # If there are no releases, set release to None
        if len(self.available_releases) == 0:
            self._release = None

        # If there is only one release, pick it
        elif len(self.available_releases) == 1:
            self._release = self.available_releases[0]

        # If there are multiple releases available, try to pick the one that matches
        # the most file metadata (album and track number)
        elif len(self.available_releases) > 1:
            rels = self.available_releases.copy()

            ## First, sort the releases:

            # Sort from oldest to newest. We append "zz-zz" to year-only releasedates in the sorting key
            # to make sure that full release dates take precedence.
            rels.sort(
                key=lambda rel: (
                    rel.releasedate if len(rel.releasedate) > 4 else rel.releasedate + "-zz-zz"
                )
            )

            # Prefer albums, then EPs, then singles, then others.
            sort_key = {"album": 0, "ep": 1, "single": 2, "other": 3}
            rels.sort(key=lambda rel: sort_key.get(rel.group.primary_type, 4))

            # Prefer digital releases over physical releases
            sort_key = {"Digital Media": 0, "CD": 1, "Casette": 2, "Vinyl": 3}
            rels.sort(key=lambda rel: sort_key.get(rel.format, 4))

            # Move unofficial releases and compilations to the end of the list.
            rels.sort(
                key=lambda rel: int(
                    rel.status != "official" or "compilation" in rel.group.secondary_types
                )  # official = 0, unofficial = 1
            )

            _original_rels = rels.copy()

            ## If we have information about the release based on metadata, drop releases
            ## that don't match them, unless we end up with no releases

            # Check 1. Check if the album title matches
            if album:
                rels_q = [rel for rel in rels if album == rel.title]
                # Check 1.5. Check if the simplified album title matches
                if not rels_q and simplify_string(album):
                    rels_q = [
                        rel for rel in rels if simplify_string(album) == simplify_string(rel.title)
                    ]
                # These two lines ensure that we don't end up with 0 releases post-query;
                # they are present in the remaining checks as well
                if rels_q:
                    rels = rels_q

            # Check 2. Check if the release date matches
            if releasedate:
                rels_q = [rel for rel in rels if releasedate == rel.releasedate]
                # Check 2.5: Check if the release year matches
                if not rels_q and len(releasedate) >= 4:
                    rels_q = [rel for rel in rels if releasedate[:4] == rel.releasedate[:4]]
                if rels_q:
                    rels = rels_q

            # Check 3. Check if the total track number matches
            if totaltracknumber:
                rels_q = [
                    rel
                    for rel in rels
                    if (rel.totaltracknumber and totaltracknumber == rel.totaltracknumber)
                    or not rel.totaltracknumber
                ]
                if rels_q:
                    rels = rels_q

            # If we end up with no releases somehow, go back to pre-heuristic releases
            if not rels:
                rels = _original_rels

            # The first remaining release wins
            self._release = rels[0]

            # Save sorted releases to available_releases for later algorithms to re-use
            self._available_releases = rels

            del rels

    async def apply_data_to_file(self, file):
        """
        Takes an EartagFile and applies the data from this recording to it.
        """
        try:
            self.release  # noqa: B018
        except ValueError:
            print("No release")
            return False

        # Covers are downloaded by the identify process ahead of time

        for prop in (
            "title",
            "artist",
            "album",
            "genre",
            "albumartist",
            "releasedate",
            "tracknumber",
            "totaltracknumber",
        ):
            if self.get_property(prop):
                file.set_property(prop, self.get_property(prop))

        if self.props.front_cover_path:
            await file.set_cover_from_path(CoverType.FRONT, self.props.front_cover_path)

        if self.props.back_cover_path:
            await file.set_cover_from_path(CoverType.BACK, self.props.back_cover_path)

        file.props.musicbrainz_recordingid = self.recording_id
        file.props.musicbrainz_albumid = self.release.release_id
        file.props.musicbrainz_releasegroupid = self.release.group.relgroup_id
        file.props.musicbrainz_trackid = self.media.get("tracks", self.media.get("track"))[0]["id"]
        file.props.musicbrainz_artistid = self.mb_data.get(
            "artist-credit-id", self.mb_data["artist-credit"][0]["artist"]["id"]
        )

    @GObject.Property(type=str)
    def recording_id(self):
        """ID of the MusicBrainz recording."""
        return self._recording_id

    @recording_id.setter
    def recording_id(self, value):
        if self._recording_id == value:
            return
        self._recording_id = value

    @GObject.Property()
    def release(self):
        """The MusicBrainz Release for this recording."""
        if self._release == MusicBrainzRecording.SELECT_RELEASE_FIRST:
            raise ValueError("Multiple releases available, select one first")
        return self._release

    @release.setter
    def release(self, value):
        self._release = value

        # Update everything that depends on the release:
        for prop in (
            "title",
            "album",
            "albumartist",
            "genre",
            "tracknumber",
            "totaltracknumber",
            "releasedate",
            "thumbnail_path",
            "front_cover_path",
            "back_cover_path",
        ):
            self.notify(prop)

    @GObject.Property(type=str)
    def title(self):
        try:
            # Use title from selected release
            return self.release.mb_data["media"][0]["track"][0]["title"]
        except (IndexError, KeyError, TypeError):
            return self.mb_data["title"]

    @property
    def all_titles(self) -> list[str]:
        """All track titles in available releases."""
        out = []
        for rel in self.available_releases:
            try:
                out.append(rel.mb_data["media"][0]["track"][0]["title"])
            except (IndexError, KeyError, TypeError):
                pass
        return out

    @GObject.Property(type=str)
    def artist(self):
        return self.mb_data["artist-credit"][0]["name"]

    @GObject.Property(type=str)
    def album(self):
        if self.release is not None:
            return self.release.title
        return ""

    @GObject.Property(type=str)
    def albumartist(self):
        if self.release is not None:
            return self.release.artist
        return ""

    @GObject.Property(type=str)
    def genre(self):
        if self.release is not None:
            if "genres" in self.mb_data and self.mb_data["genres"]:
                return title_case_preserve_uppercase(self.mb_data["genres"][0]["name"])
            return self.release.genre
        return ""

    @property
    def media(self):
        """Shorthand for the "media" property of the current release."""
        return self.release.mb_data.get("media")[0]

    @GObject.Property(type=int)
    def tracknumber(self):
        # Weird MusicBrainz bug: In search queries it's "track", in direct
        # by-ID queries it's "tracks".
        try:
            return int(self.media.get("tracks", self.media.get("track"))[0]["number"])
        except ValueError:
            # HACK until we properly support vinyl track numbers.
            return 0

    @GObject.Property(type=int)
    def totaltracknumber(self):
        try:
            return int(self.media["track-count"])
        except ValueError:
            # HACK until we properly support vinyl track numbers.
            return 0

    @GObject.Property(type=str)
    def releasedate(self):
        if self.release is not None:
            return self.release.releasedate
        return ""

    @GObject.Property(type=str)
    def disambiguation(self):
        return self.mb_data.get("disambiguation", "")

    @GObject.Property(type=str)
    def thumbnail_path(self):
        if self.release is not None:
            return self.release.thumbnail_path
        return ""

    @GObject.Property(type=str)
    def front_cover_path(self):
        if self.release is not None:
            return self.release.front_cover_path
        return ""

    @GObject.Property(type=str)
    def back_cover_path(self):
        if self.release is not None:
            return self.release.back_cover_path
        return ""

    async def download_covers_async(self):
        """Downloads the covers for the release from coverartarchive.org"""
        return await self.release.download_covers_async()

    def __str__(self):
        if self.disambiguation:
            return f"MusicBrainzRecording {self.recording_id} ({self.title} - {self.artist} ({self.disambiguation}))"
        return f"MusicBrainzRecording {self.recording_id} ({self.title} - {self.artist})"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if not isinstance(other, MusicBrainzRecording):
            return False
        return self.recording_id == other.recording_id

    def __hash__(self):
        return hash(self.recording_id)


class MusicBrainzRelease(GObject.Object):
    __gtype_name__ = "MusicBrainzRelease"

    NEED_UPDATE_COVER = -2

    cover_cache = {}

    def __init__(self, release_data):
        super().__init__()
        self.mb_data = release_data

        self.thumbnail_dl_task = None

        self.group = MusicBrainzReleaseGroup(self.mb_data["release-group"])

        self.thumbnail = EartagCAACover("release", self.release_id, "thumbnail")
        self.front_cover = EartagCAACover("release", self.release_id, "front")
        self.back_cover = EartagCAACover("release", self.release_id, "back")

    @staticmethod
    async def new_for_id(id):
        data = await mb_query.download(
            build_url(
                "release",
                id,
                inc=[
                    "artist-credits",
                    "recordings",
                    "release-groups",
                    "genres",
                    "media",
                ],
            )
        )
        return MusicBrainzRelease(data)

    @GObject.Property(type=str)
    def release_id(self):
        return self.mb_data["id"]

    @GObject.Property(type=str)
    def title(self):
        return self.mb_data["title"]

    @GObject.Property(type=str)
    def artist(self):
        return self.mb_data["artist-credit"][0]["name"]

    @GObject.Property(type=str)
    def genre(self):
        if "genres" in self.mb_data and self.mb_data["genres"]:
            return title_case_preserve_uppercase(self.mb_data["genres"][0]["name"])
        return ""

    @GObject.Property(type=str)
    def releasedate(self):
        if "date" in self.mb_data:
            return self.mb_data["date"]
        if "first-release-date" in self.group.mb_data:
            return self.group.mb_data["first-release-date"]
        return ""

    @GObject.Property(type=int)
    def totaltracknumber(self):
        return int(self.mb_data["media"][0]["track-count"])

    @GObject.Property(type=str)
    def status(self):
        if "status" in self.mb_data and self.mb_data["status"]:
            return self.mb_data["status"].lower()
        return "official"

    @GObject.Property(type=str)
    def format(self):
        return self.mb_data["media"][0].get("format", "")

    @GObject.Property(type=str)
    def disambiguation(self):
        return "; ".join(
            [
                x
                for x in (
                    self.props.format,
                    self.mb_data.get("disambiguation", None),
                )
                if x
            ]
        )

    # Covers

    @GObject.Property(type=str)
    def thumbnail_path(self):
        if not self.thumbnail.props.path:
            return self.group.thumbnail_path
        return self.thumbnail.props.path

    @GObject.Property(type=bool, default=False)
    def thumbnail_loaded(self):
        return self.thumbnail.props.loaded

    @GObject.Property(type=str)
    def front_cover_path(self):
        if not self.front_cover.props.loaded:
            raise ValueError("Covers have not been downloaded yet; run download_covers_async()")
        if not self.front_cover.props.path:
            return self.group.front_cover_path
        return self.front_cover.props.path

    @GObject.Property(type=str)
    def back_cover_path(self):
        if not self.back_cover.props.loaded:
            raise ValueError("Covers have not been downloaded yet; run download_covers_async()")
        return self.back_cover.props.path

    def queue_thumbnail_download(self):
        """Queues a download of the thumbnail for the release from coverartarchive.org"""
        if (
            self.thumbnail_dl_task
            and not self.thumbnail.props.loaded
            and not self.thumbnail_dl_task.done()
        ):
            self.thumbnail_dl_task.cancel()
        self.thumbnail_dl_task = event_loop.create_task(self.download_thumbnail_async())

    async def download_thumbnail_async(self):
        """Downloads the covers for the release from coverartarchive.org"""
        await self.thumbnail.download()
        if not self.thumbnail.props.loaded:
            await self.group.thumbnail.download()
        self.notify("thumbnail-path")
        self.notify("thumbnail-loaded")

    async def download_covers_async(self):
        """Downloads the covers for the release from coverartarchive.org"""
        await self.front_cover.download()
        if not self.front_cover.props.path:
            await self.group.front_cover.download()
        self.notify("front-cover-path")
        await self.back_cover.download()
        self.notify("back-cover-path")

    def __str__(self):
        return f"MusicBrainzRelease {self.release_id} ({self.title} - {self.artist} ({self.disambiguation}))"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if not isinstance(other, MusicBrainzRelease):
            return False
        return self.release_id == other.release_id

    def __hash__(self):
        return hash(self.release_id)


class MusicBrainzReleaseGroup(GObject.Object):
    """A container for release group information, as found in the release query."""

    __gtype_name__ = "MusicBrainzReleaseGroup"

    NO_COVER = -2

    def __init__(self, relgroup_data):
        super().__init__()
        self._releases = None

        self.mb_data = relgroup_data

        self.thumbnail = EartagCAACover("release-group", self.relgroup_id, "thumbnail")
        self.front_cover = EartagCAACover("release-group", self.relgroup_id, "front")

    @staticmethod
    async def _fetch_full_data(groupid):
        return await mb_query.download(build_url("release-group", groupid, inc=["releases"]))

    @staticmethod
    async def new_for_id(id):
        data = await MusicBrainzReleaseGroup._fetch_full_data(id)
        return MusicBrainzReleaseGroup(relgroup_data=data)

    @GObject.Property(type=str)
    def relgroup_id(self):
        return self.mb_data["id"]

    @GObject.Property(type=str)
    def primary_type(self):
        try:
            return self.mb_data["primary-type"].lower()
        except AttributeError:
            return "other"

    @GObject.Property(type=str)
    def secondary_types(self):
        try:
            return [t.lower() for t in self.mb_data.get("secondary-types", [])]
        except AttributeError:
            return []

    @GObject.Property
    def release_ids(self):
        return [r["id"] for r in self.mb_data["releases"]]

    @GObject.Property
    def releases(self):
        if not self._releases:
            raise ValueError("Releases have not been initialized yet; run get_releases_async()")
        return self._releases

    async def get_releases_async(self):
        if not self._releases:
            self._releases = [await MusicBrainzRelease.new_for_id(id) for id in self.release_ids]

    @GObject.Property(type=str)
    def thumbnail_path(self):
        return self.thumbnail.props.path

    @GObject.Property(type=str)
    def front_cover_path(self):
        if not self.front_cover.props.loaded:
            raise ValueError("Covers have not been downloaded yet; run download_covers_async()")
        return self.front_cover.props.path

    def __eq__(self, other):
        if not isinstance(other, MusicBrainzReleaseGroup):
            return False
        return self.relgroup_id == other.relgroup_id

    def __hash__(self):
        return hash(self.relgroup_id)


async def acoustid_identify_file(file) -> Tuple[float, "MusicBrainzRecording"]:
    """
    Uses AcoustID and Chromaprint to identify a track's data.

    Returns a tuple containing the confidence and MusicBrainzRecording
    object for the file, or (0.0, None) if it couldn't be found.
    """
    logger.debug(f"Running AcoustID identification for file {os.path.basename(file.path)}")

    try:
        results = await asyncio.to_thread(acoustid.match, ACOUSTID_API_KEY, file.path, parse=False)
        if "results" not in results or not results["results"]:
            return (0.0, None)
    except:  # noqa: E722
        logger.warning(
            f"Error while getting AcoustID match for {os.path.basename(file.path)} ({file.id}):"
        )
        traceback.print_exc()
        logger.warning("Continuing without match. (This is not a fatal error!)")
        return (0.0, None)

    acoustid_data = results["results"][0]

    logger.debug(f"AcoustID identification result: {acoustid_data['score']}")

    if acoustid_data["score"] * 100 < config["acoustid-confidence-treshold"]:
        return (0.0, None)

    if "recordings" in acoustid_data:
        musicbrainz_id = acoustid_data["recordings"][0]["id"]
        logger.debug(f"AcoustID matched recording is {musicbrainz_id}")
        rec = await MusicBrainzRecording.new_for_id(musicbrainz_id)
        rec.sort_releases()  # Sort releases in the recording
        return (acoustid_data["score"], rec)

    logger.debug("No recordings in AcoustID result")
    return (0.0, None)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from .backends import FILE_CLASSES
from .backends.file import EartagFile, BASIC_TAGS, EXTRA_TAGS, TAG_NAMES, CoverType
from .logger import logger
from .utils import get_readable_length, file_is_sandboxed
from .utils.validation import is_valid_image_file
from .utils.widgets import EartagAlbumCoverImage, EartagPopoverButton  # noqa: F401
from .tagentry import (  # noqa: F401
    EartagTagEntryManager,
    EartagTagEntry,
    EartagTagEntryRow,
    EartagTagEditableLabel,
)
from . import APP_GRESOURCE_PATH

from gi.repository import Adw, Gtk, Gdk, Gio, GLib, GObject

import asyncio
from functools import cached_property
import mimetypes
import traceback
import os.path

from collections.abc import Iterable


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/albumcoverbutton.ui")
class EartagAlbumCoverButton(Adw.Bin):
    __gtype_name__ = "EartagAlbumCoverButton"

    button = Gtk.Template.Child()
    cover_image = Gtk.Template.Child()

    highlight_revealer = Gtk.Template.Child()
    highlight_stack = Gtk.Template.Child()
    drop_highlight = Gtk.Template.Child()
    hover_highlight = Gtk.Template.Child()

    handling_drag = False
    handling_undefined_drag = False
    image_file_filter = Gtk.Template.Child()

    front_toggle = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        self._cover_type = CoverType.FRONT
        self._remove_undo_buffer = {}
        self.cover_tempdir = None

        self.connect("destroy", self.on_destroy)
        self.drop_target = Gtk.DropTarget(
            actions=Gdk.DragAction.COPY,
            formats=Gdk.ContentFormats.new_for_gtype(Gio.File),
        )

        self.drop_target.connect("accept", self.on_drag_accept)
        self.drop_target.connect("enter", self.on_drag_hover)
        self.drop_target.connect("leave", self.on_drag_unhover)
        self.drop_target.connect("drop", self.on_drag_drop)
        self.add_controller(self.drop_target)

        self.hover_controller = Gtk.EventControllerMotion.new()
        self.hover_controller.connect("enter", self.on_hover)
        self.hover_controller.connect("leave", self.on_unhover)
        self.add_controller(self.hover_controller)

        self.front_toggle.connect("notify::active", self.update_from_switcher)
        self.front_toggle.set_active(True)

        self.cover_image.connect("cover-changed", self.update_coverbutton_save_availability)
        self.cover_image.connect("notify::cover-type", self.update_coverbutton_save_availability)
        self.cover_image.connect("notify::is-empty", self.update_coverbutton_save_availability)

        self.bind_property("cover-type", self.cover_image, "cover-type")

        # Register actions for popover menu
        self.install_action("albumcoverbutton.load", None, self.show_cover_file_chooser)
        self.install_action("albumcoverbutton.save", None, self.save_cover)
        self.action_set_enabled("albumcoverbutton.save", False)
        self.install_action("albumcoverbutton.remove", None, self.remove_cover)

        self.files = set()

    def _dropdown_lockup_workaround(self, toggle, *args):
        """
        Workaround for https://gitlab.gnome.org/GNOME/gtk/-/issues/5568
        """
        if not toggle.get_active():
            self.button.popover.popdown()

    def update_from_switcher(self, toggle, *args):
        """Sets the displayed cover by checking the cover switcher."""
        if toggle.get_active():
            self.cover_type = CoverType.FRONT
        else:
            self.cover_type = CoverType.BACK

    @GObject.Property(type=int)
    def cover_type(self):
        """Whether to display the front or back cover."""
        return self._cover_type

    @cover_type.setter
    def cover_type(self, value):
        self._cover_type = value
        if self.cover_type == CoverType.FRONT:
            title = _("Album cover (front)")
        else:
            title = _("Album cover (back)")

        Gtk.Accessible.update_property(self, (Gtk.AccessibleProperty.LABEL,), (title,))
        Gtk.Accessible.update_property(self.button, (Gtk.AccessibleProperty.LABEL,), (title,))
        self.button.set_tooltip_text(title)

    def update_coverbutton_save_availability(self, *args):
        if self.cover_type == CoverType.FRONT:
            cover = "front_cover"
        else:
            cover = "back_cover"

        self.action_set_enabled("albumcoverbutton.save", not self.cover_image.is_empty)

        enable_remove = False
        for file in self.files:
            if getattr(file, cover):
                enable_remove = True
                break

        self.action_set_enabled("albumcoverbutton.remove", enable_remove)

    def bind_to_files(self, files):
        self.files |= set(files)

        file = next(iter(self.files))
        if len(self.files) < 2:
            if not file.supports_album_covers:
                self.set_visible(False)
                self.update_coverbutton_save_availability()
                return False
            else:
                self.set_visible(True)
            self.cover_image.bind_to_file(file)
            self.cover_image.mark_as_nonempty()
        else:
            covers_different = False
            our_cover = file.get_cover(self.cover_type)

            if False in [f.supports_album_covers for f in self.files]:
                self.set_visible(False)
            else:
                self.set_visible(True)

            for _file in self.files:
                if _file.get_cover(self.cover_type) != our_cover:
                    covers_different = True
                    self.cover_image.mark_as_empty()
                    break
            if not covers_different:
                self.cover_image.mark_as_nonempty()

    def unbind_from_files(self, files):
        for file in files:
            if file in self.files:
                self.files.remove(file)

        for _file in self.files:
            if not _file.supports_album_covers:
                self.set_visible(False)
                self.update_coverbutton_save_availability()
                break
            else:
                self.set_visible(True)

        if len(self.files) > 0:
            file = next(iter(self.files))
            if len(self.files) > 1:
                covers_different = False
                our_cover = file.get_cover(self.cover_type)
                for _file in self.files:
                    if _file.get_cover(self.cover_type) != our_cover:
                        covers_different = True
                        self.cover_image.mark_as_empty()
                        break
                if not covers_different:
                    self.cover_image.mark_as_nonempty()
                    if file.supports_album_covers and file.get_cover(self.cover_type):
                        self.cover_image.bind_to_file(file)

            elif len(self.files) == 1:
                self.cover_image.bind_to_file(file)
                self.cover_image.on_cover_change()

    def on_destroy(self, *args):
        self.files = None
        if self.cover_tempdir:
            self.cover_tempdir.cleanup()
            self.cover_tempdir = None

    def show_cover_file_chooser(self, *args):
        asyncio.create_task(self.show_cover_file_chooser_async())

    async def show_cover_file_chooser_async(self):
        """Shows the file chooser."""
        file_chooser = Gtk.FileDialog(title=_("Select Album Cover Image"), modal=True)

        _filters = Gio.ListStore.new(Gtk.FileFilter)
        _filters.append(self.image_file_filter)
        file_chooser.set_filters(_filters)

        try:
            gfile = await file_chooser.open(self.get_native(), None)
        except GLib.GError:
            return

        if not gfile:
            return

        for file in self.files:
            try:
                await file.set_cover_from_path(self.cover_type, gfile.get_path())
            except:  # noqa: E722
                logger.error(f"Error while setting cover for file {file}:")
                traceback.print_exc()

        self.cover_image.on_cover_change()

    def save_cover(self, *args):
        asyncio.create_task(self.save_cover_async())

    async def save_cover_async(self):
        """Opens a file dialog to have the cover art to a file."""
        if self.cover_type == CoverType.FRONT:
            cover = self.files[0].front_cover
        elif self.cover_type == CoverType.BACK:
            cover = self.files[0].back_cover
        else:
            return

        cover_extension = mimetypes.guess_extension(cover.mime)
        target_folder, target_filename = os.path.split(self.files[0].path)
        target_filename = os.path.splitext(target_filename)[0] + cover_extension

        file_chooser = Gtk.FileDialog(
            title=_("Save Album Cover To…"),
            modal=True,
            initial_folder=Gio.File.new_for_path(target_folder),
            initial_name=target_filename,
        )

        try:
            response = await file_chooser.save(self.get_native(), None)
        except GLib.GError:
            return

        if not response:
            return

        save_path = response.get_path()

        if cover_extension and not save_path.endswith(cover_extension):
            save_path += cover_extension

        try:
            await cover.save_to_path(save_path)
        except:  # noqa: E722
            logger.error("Failed to save cover:")
            traceback.print_exc()
        else:
            # TRANSLATORS: {path} is a placeholder for the path.
            # **Do not change the text between the curly brackets!**
            toast = Adw.Toast.new(_("Saved cover to {path}").format(path=save_path))
            self.get_native().toast_overlay.add_toast(toast)

    def remove_cover(self, *args):
        self._remove_undo_buffer = {}
        self._remove_undo_buffer["type"] = self.cover_type

        if self.cover_type == CoverType.FRONT:
            cover_prop = "front_cover"
        elif self.cover_type == CoverType.BACK:
            cover_prop = "back_cover"

        for file in self.files:
            cover = file.get_property(cover_prop)
            cover_is_modified = cover_prop in file.modified_tags
            self._remove_undo_buffer[file.id] = [
                cover,
                cover_is_modified,
            ]
            file.delete_cover(self.cover_type)

        self.cover_image.on_cover_change()

        remove_msg = ngettext(
            "Removed cover from file", "Removed covers from {n} files", len(self.files)
        ).format(n=len(self.files))
        toast = Adw.Toast.new(remove_msg)
        toast.set_button_label(_("Undo"))
        toast.connect("button-clicked", self._remove_undo)
        toast.connect("dismissed", self._remove_undo_clear)
        self.get_native().toast_overlay.add_toast(toast)

    def _remove_undo(self, *args):
        asyncio.create_task(self._remove_undo_async())

    async def _remove_undo_async(self):
        if self._remove_undo_buffer["type"] == CoverType.FRONT:
            cover_prop = "front_cover"
        elif self._remove_undo_buffer["type"] == CoverType.BACK:
            cover_prop = "back_cover"
        else:
            return

        file_manager = self.get_native().file_manager
        for file in file_manager.files:
            if file.id not in self._remove_undo_buffer:
                continue
            file.set_property(cover_prop, self._remove_undo_buffer[file.id][0])
            was_modified = self._remove_undo_buffer[file.id][1]
            if was_modified:
                file.mark_as_modified(cover_prop)
            else:
                file.mark_tag_as_unmodified(cover_prop)

        self.cover_image.on_cover_change()

        self._remove_undo_clear()

    def _remove_undo_clear(self, *args):
        if "type" not in self._remove_undo_buffer:
            return

        for k, v in self._remove_undo_buffer.items():
            if k == "type":
                continue
            del v[0]  # Make absolutely sure the covers are unreffed
        self._remove_undo_buffer = {}

    # Drag-and-drop

    def on_drag_accept(self, target, drop, *args):
        drop.read_value_async(Gio.File, 0, None, self.verify_file_valid)
        return True

    def verify_file_valid(self, drop, task, *args):
        file = drop.read_value_finish(task)
        path = file.get_path()
        if not is_valid_image_file(path):
            self.handling_undefined_drag = True
            self.drop_target.reject()
            self.on_drag_unhover()
        else:
            self.handling_undefined_drag = False

    def on_drag_hover(self, *args):
        self.handling_drag = True
        self.highlight_stack.set_visible_child(self.drop_highlight)
        self.highlight_revealer.set_reveal_child(True)
        return Gdk.DragAction.COPY

    def on_drag_unhover(self, *args):
        self.highlight_revealer.set_reveal_child(False)
        self.handling_drag = False

    def on_drag_drop(self, drop_target, value, *args):
        asyncio.create_task(self.on_drag_drop_async(drop_target, value))

    async def on_drag_drop_async(self, drop_target, value):
        path = value.get_path()
        if not path:
            return
        for file in self.files:
            await file.set_cover_from_path(self.cover_type, path)
        self.cover_image.on_cover_change()
        self.on_drag_unhover()

    # Hover
    def on_hover(self, *args):
        if not self.handling_drag and not self.handling_undefined_drag:
            self.highlight_stack.set_visible_child(self.hover_highlight)
            self.highlight_revealer.set_reveal_child(True)

    def on_unhover(self, *args):
        self.highlight_revealer.set_reveal_child(False)


extra_tag_names = dict([(k, v) for k, v in TAG_NAMES.items() if k in ["none"] + list(EXTRA_TAGS)])
extra_tag_names_swapped = dict([(v, k) for k, v in extra_tag_names.items()])
more_item_tag_strings = Gtk.StringList.new(list(extra_tag_names.values()))


class EartagExtraTagRow(EartagTagEntryRow):
    __gtype_name__ = "EartagExtraTagRow"

    def __init__(self, tag, parent):
        super().__init__()
        self.props.visible = False
        self.parent = parent

        self.bound_property = tag

        self.is_numeric = tag in EartagFile.int_properties
        self.is_float = tag in EartagFile.float_properties

        self.row_remove_button = Gtk.Button(
            icon_name="edit-delete-symbolic", valign=Gtk.Align.CENTER
        )
        self.row_remove_button.props.tooltip_text = _("Remove tag")
        self.row_remove_button.add_css_class("flat")
        self._remove_clicked_connect = self.row_remove_button.connect(
            "clicked", self.remove_button_pressed
        )
        self.add_suffix(self.row_remove_button)

        self._destroy_connect = self.connect("destroy", self._on_destroy)

    @GObject.Property(type=str, default="")
    def bound_property(self):
        try:
            return self._bound_property
        except AttributeError:
            return ""

    @bound_property.setter
    def bound_property(self, value: str):
        self._bound_property = value
        self.set_title(extra_tag_names[value])

    def remove_button_pressed(self, *args):
        self.parent.remove_tag(self.bound_property)

    def _on_destroy(self, *args):
        try:
            self.row_remove_button  # noqa: B018
        except AttributeError:
            return
        self.row_remove_button.disconnect(self._remove_clicked_connect)
        self.disconnect(self._destroy_connect)
        del self.row_remove_button
        del self.parent


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/moretagsgroup.ui")
class EartagMoreTagsGroup(Gtk.Box):
    """
    Used for the "More tags" row in the FileView.
    """

    __gtype_name__ = "EartagMoreTagsGroup"

    tag_entry_listbox = Gtk.Template.Child()
    tag_selector = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

        self.tagentry_manager = EartagTagEntryManager()

        #: Filetypes bound to this entry (gtype name: count).
        self.filetypes: dict[str, int] = {}

        #: Files bound to this entry.
        self.files: set[EartagFile] = set()

        #: File connections.
        self.file_connections: dict[str, list[int]] = {}

        # Set up allowed tag lists for each filetype. Since different filetypes
        # have different sets of allowed extra tags, we store them in this class
        # and use this information to show/hide certain entries.
        self.allowed_tags_for_filetype: dict[str, Iterable[str]] = {}
        for filetype in FILE_CLASSES:
            self.allowed_tags_for_filetype[filetype.__gtype_name__] = filetype.supported_extra_tags

        #: Entry rows under this group.
        self.entries: dict[str, Gtk.Widget] = {}

        #: All present properties in files, and the IDs of the files that have them.
        #: To get the actually shown properties, check self.tagentry_manager.managed_properties
        #: (since that value excludes blocked tags).
        self.present_props: dict[str, set[str]] = {}

        self.tag_filter = Gtk.CustomFilter.new(self.tag_filter_func, self.tag_selector.tag_model)
        self.tag_selector.set_filter(self.tag_filter)

        # Set up the entries for all extra tags. The entries start out hidden;
        # once they become needed, we show them again. This allows us to
        # avoid having to create/destroy the row objects every time, and is
        # relatively cheap since all entry-related operations are only
        # done in the manager, and only bound to it when shown.

        for tag in EXTRA_TAGS:
            self.entries[tag] = EartagExtraTagRow(tag, self)
            self.tag_entry_listbox.append(self.entries[tag])

    @cached_property
    def blocked_tags(self) -> set[str]:
        """List of tags not allowed for all available filetypes."""
        allowed_types_values = []
        for ftype, values in self.allowed_tags_for_filetype.items():
            if self.filetypes.get(ftype, 0) > 0:
                allowed_types_values.append(values)

        if allowed_types_values:
            allowed_types_intersection = set.intersection(*[set(x) for x in allowed_types_values])
        else:
            allowed_types_intersection = set(EXTRA_TAGS)

        return set(EXTRA_TAGS) - allowed_types_intersection

    # File management

    async def bind_to_files(self, files: Iterable[EartagFile]):
        """Bind to the files in the provided iterable.

        Remember to call self."""
        old_filetypes = set(k for k, v in self.filetypes.items() if v)
        for file in files:
            if file in self.files:
                continue
            self.files.add(file)
            self.file_connections[file.id] = [file.connect("modified", self.on_file_modified)]

            present_extra_tags = set(file.present_extra_tags)

            for tag in present_extra_tags:
                if tag in self.present_props:
                    self.present_props[tag].add(file.id)
                    if len(self.present_props[tag]) == 1:
                        self.show_entry(tag)
                else:
                    self.present_props[tag] = set((file.id,))
                    self.show_entry(tag)

            if file.__gtype_name__ in self.filetypes:
                self.filetypes[file.__gtype_name__] += 1
            else:
                self.filetypes[file.__gtype_name__] = 1
        new_filetypes = set(k for k, v in self.filetypes.items() if v)

        if old_filetypes != new_filetypes:
            # Clear cached blocked tags value
            try:
                del self.blocked_tags
            except AttributeError:
                pass

            # Hide blocked entries
            for tag in set.intersection(
                self.blocked_tags, self.tagentry_manager.managed_properties
            ):
                self.hide_entry(tag)

        await self.tagentry_manager.bind_to_files(files)

    async def unbind_from_files(self, files: Iterable[EartagFile]):
        """Unbind from the files in the provided iterable."""
        old_filetypes = set(k for k, v in self.filetypes.items() if v)
        for file in files:
            if file not in self.files:
                continue
            self.files.remove(file)
            for conn in self.file_connections[file.id]:
                file.disconnect(conn)

            present_extra_tags = set(file.present_extra_tags)

            for tag in present_extra_tags:
                self.present_props[tag].remove(file.id)
                if len(self.present_props[tag]) == 0:
                    del self.present_props[tag]

            self.filetypes[file.__gtype_name__] -= 1
        new_filetypes = set(k for k, v in self.filetypes.items() if v)

        if old_filetypes != new_filetypes:
            old_blocked_tags = self.blocked_tags.copy()

            # Clear cached blocked tags value
            try:
                del self.blocked_tags
            except AttributeError:
                pass

            # Unhide blocked entries
            for tag in old_blocked_tags - self.blocked_tags:
                if tag in self.present_props:
                    self.show_entry(tag)

        await self.tagentry_manager.unbind_from_files(files)

    # Entry management

    def show_entry(self, prop: str):
        """Show the extra row for the given property."""
        self.entries[prop].props.visible = True
        self.tagentry_manager.add_entry(prop, self.entries[prop])
        self.refresh_tag_filter()

    def hide_entry(self, prop: str):
        """Hide the extra row for the given property."""
        self.entries[prop].props.visible = False
        self.tagentry_manager.entry_inconsistency[prop] = False
        self.tagentry_manager.remove_entry(prop)
        self.refresh_tag_filter()

    def remove_tag(self, prop: str):
        """Remove the tag from all open files."""
        for file in self.tagentry_manager.files:
            if prop in file.present_extra_tags:
                file.present_extra_tags.remove(prop)
                file.delete_tag(prop)

    def on_file_modified(self, file: EartagFile, tag: str):
        """Handle a file being modified, in case an extra tag is added/removed."""
        if tag not in EXTRA_TAGS:
            return
        if tag in file.present_extra_tags:
            if tag in self.present_props:
                self.present_props[tag].add(file.id)
            else:
                self.present_props[tag] = set([file.id])
        else:
            if tag in self.present_props and file.id in self.present_props[tag]:
                self.present_props[tag].remove(file.id)
                if len(self.present_props[tag]) == 0:
                    del self.present_props[tag]

        # The ignore_file_change check here makes sure that we don't remove an
        # entry when it's cleared by the user by removing the text.
        if not self.tagentry_manager._ignore_file_change:
            self.refresh_entry_visibility_for_tag(tag)

    def refresh_entry_visibility(self):
        """
        Synchronize entry state with self.present_props.

        There are two cases where the state is desynced:

        - During binding/unbinding, we don't show/hide entries immediately;
          this is a bit of an ugly hack to prevent the entry from getting cleared
          then filled again during the unbind/bind cycle, which causes a
          distracting animation.

        - The state can get desynchronized if the user adds a tag entry, types
          something in, then removes it; in those cases, we can't immediately
          hide the entry. At the same time, we must hide the entry if it is
          either deleted manually or the underlying tag disappears because of
          an undo, redo or delete all tags operation.
        """
        target_props = set(self.present_props.keys()) - self.blocked_tags
        current_props = self.tagentry_manager.managed_properties

        # Entries to show:
        for tag in target_props - current_props:
            self.show_entry(tag)

        # Entries to hide:
        for tag in current_props - target_props:
            self.hide_entry(tag)

    def refresh_entry_visibility_for_tag(self, tag: str):
        """Same as refresh_entry_visibility, but for a single tag."""
        target_props = set(self.present_props.keys()) - self.blocked_tags
        current_props = self.tagentry_manager.managed_properties

        if tag in target_props and tag not in current_props:
            self.show_entry(tag)
        elif tag not in target_props and tag in current_props:
            self.hide_entry(tag)

    # Tag selection dropdown

    @Gtk.Template.Callback()
    def add_row_from_selector(self, selector, tag):
        """Adds a new row based on the tag selector."""
        self.show_entry(tag)
        self.present_props[tag] = set(f.id for f in self.files if f.has_tag(tag))
        self.entries[tag].grab_focus()

    def refresh_tag_filter(self, *args):
        """Refreshes the filter for the additional tag add row."""
        self.tag_filter.changed(Gtk.FilterChange.DIFFERENT)
        self.tag_entry_listbox.props.visible = bool(self.tagentry_manager.managed_properties)

    def tag_filter_func(self, _tag_name, *args):
        """Filter function for the tag dropdown."""
        if not self.tag_selector.tag_filter_func(_tag_name):
            return False

        present_tags = self.tagentry_manager.managed_properties

        tag_name = _tag_name.get_string()
        tag_prop = self.tag_selector.tag_names_swapped[tag_name]

        if tag_prop in ("length", "bitrate") + BASIC_TAGS:
            return False
        if tag_prop == "none":
            return False
        if tag_prop in present_tags:
            return False
        if tag_prop in self.blocked_tags:
            return False

        return True

    # Miscelaneous

    @GObject.Property(type=bool, default=False)
    def height_below_360(self):
        """Hack to fix https://gitlab.gnome.org/World/eartag/-/merge_requests/130"""
        return self._height_below_360

    @height_below_360.setter
    def height_below_360(self, value):
        self._height_below_360 = value
        if value is True:
            self.tag_selector.set_direction(Gtk.ArrowType.UP)
        else:
            self.tag_selector.set_direction(Gtk.ArrowType.DOWN)


class EartagFileInfoLabel(Gtk.Label):
    """Label showing information about opened files."""

    __gtype_name__ = "EartagFileInfoLabel"

    def __init__(self):
        super().__init__()
        self.add_css_class("dim-label")
        self.add_css_class("numeric")
        self.files = set()
        self.refresh_label()

    def bind_to_files(self, files):
        self.files |= set(files)
        self.refresh_label()

    def unbind_from_files(self, files):
        self.files -= set(files)
        self.refresh_label()

    def refresh_label(self):
        if len(self.files) == 0:
            self.set_label("")
        elif len(self.files) == 1:
            self._set_info_label(next(iter(self.files)))
        else:
            self.set_label(_("(Multiple files selected)"))

    def _set_info_label(self, file):
        length_readable = get_readable_length(int(file.length))

        # Get human-readable version of channel count
        channels = file.channels
        if channels == 0:
            channels_readable = "N/A"
        elif channels == 1:
            channels_readable = "Mono"
        elif channels == 2:
            channels_readable = "Stereo"
        else:
            channels_readable = ngettext("1 channel", "{n} channels", channels).format(n=channels)  # noqa: E501

        if file.bitrate > -1:
            bitrate_readable = str(file.bitrate)
        else:
            bitrate_readable = "N/A"

        self.set_label(
            "{length} • {bitrate} kbps • {channels} • {filetype}".format(
                filetype=file.filetype,
                length=length_readable,
                bitrate=bitrate_readable,
                channels=channels_readable,
            )
        )


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/filenamerow.ui")
class EartagFilenameRow(Adw.EntryRow):
    __gtype_name__ = "EartagFilenameRow"

    error = GObject.Property(type=bool, default=False)

    def __init__(self):
        super().__init__()
        self.files = set()
        self._connections = {}
        self._title = self.props.title
        self.get_delegate().connect("insert-text", self.validate_input)

    def bind_to_files(self, files: Iterable[EartagFile]):
        self.files |= set(files)
        for file in files:
            self._connections[file.id] = file.connect("notify::path", self.update_on_bind)
        self.update_on_bind()

    def unbind_from_files(self, files: Iterable[EartagFile]):
        for file in files:
            if file in self.files:
                file.disconnect(self._connections[file.id])
                del self._connections[file.id]
                self.files.remove(file)
        self.update_on_bind()

    def update_on_bind(self, *args):
        if len(self.files) > 1:
            self.props.title = self._title + " " + _("(multiple files)")
            self.set_editable(False)
            self.props.show_apply_button = False
            self.set_text("")
        elif len(self.files) == 1:
            path = next(iter(self.files)).path
            self.props.title = self._title
            self.set_editable(not file_is_sandboxed(path))
            self.props.show_apply_button = True
            self.set_text(os.path.basename(path))
        else:
            self.props.title = self._title
            self.set_editable(False)
            self.props.show_apply_button = False
            self.set_text("")

    @Gtk.Template.Callback()
    def set_filename(self, *args):
        if len(self.files) != 1:
            return
        file = next(iter(self.files))
        old_path = file.path
        self.get_native().file_manager.rename_files(
            (file,),
            (os.path.join(os.path.dirname(old_path), self.get_text()),),
        )

    def validate_input(self, entry, text, length, position, *args):
        if "/" in text:
            GObject.signal_stop_emission_by_name(entry, "insert-text")


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/fileview.ui")
class EartagFileView(Gtk.Stack):
    __gtype_name__ = "EartagFileView"

    loading = Gtk.Template.Child()
    content_stack = Gtk.Template.Child()
    content_scroll = Gtk.Template.Child()
    content_clamp = Gtk.Template.Child()
    select_file = Gtk.Template.Child()

    important_data_container = Gtk.Template.Child()
    tag_list = Gtk.Template.Child()

    album_cover = Gtk.Template.Child()
    title_entry = Gtk.Template.Child()
    artist_entry = Gtk.Template.Child()
    file_info = Gtk.Template.Child()
    tracknumber_entry = Gtk.Template.Child()
    totaltracknumber_entry = Gtk.Template.Child()
    album_entry = Gtk.Template.Child()
    albumartist_entry = Gtk.Template.Child()
    genre_entry = Gtk.Template.Child()
    releasedate_entry = Gtk.Template.Child()
    comment_entry = Gtk.Template.Child()
    filename_entry = Gtk.Template.Child()
    more_tags_group = Gtk.Template.Child()

    previous_file_button_revealer = Gtk.Template.Child()
    next_file_button_revealer = Gtk.Template.Child()
    previous_file_button = Gtk.Template.Child()
    next_file_button = Gtk.Template.Child()

    def __init__(self):
        """Initializes the EartagFileView."""
        super().__init__()

        self.writable = False
        self.files = set()

        self.bindable_entries = (
            self.album_cover,
            self.filename_entry,
            self.file_info,
        )

        self.tagentries = (
            self.title_entry,
            self.artist_entry,
            self.tracknumber_entry,
            self.totaltracknumber_entry,
            self.album_entry,
            self.albumartist_entry,
            self.genre_entry,
            self.releasedate_entry,
            self.comment_entry,
        )

        self.tagentry_manager = EartagTagEntryManager()

        for tagentry in self.tagentries:
            self.tagentry_manager.add_entry(tagentry.bound_property, tagentry)

        self.previous_fileview_width = 0

    def on_close(self):
        self.album_cover.on_destroy()
        self.tagentry_manager.destroy()

    def set_file_manager(self, file_manager):
        self.file_manager = file_manager
        self.file_manager.connect("refresh-needed", self.update_binds)
        self.file_manager.connect("selection-changed", self.update_binds)
        self.file_manager.load_task.connect("notify::progress", self.update_loading)
        self.file_manager.connect("notify::has-error", self.update_error)

        window = self.get_native()
        self.next_file_button.connect("clicked", window.select_next)
        self.previous_file_button.connect("clicked", window.select_previous)
        window.connect("notify::selection-mode", self.update_buttons)

    def update_error(self, *args):
        # Currently this is only used by the releasedate entry. Expand this
        # when needed.
        if self.file_manager.has_error:
            self.releasedate_entry.add_css_class("error")
        else:
            self.releasedate_entry.remove_css_class("error")

    def update_loading(self, task, *args):
        if task.progress == 0:
            self.set_visible_child(self.content_stack)
        else:
            self.set_visible_child(self.loading)

    def update_buttons(self, *args):
        """Updates the side switcher button state."""
        if len(self.file_manager.files) == 0 or self.get_native().selection_mode:
            self.previous_file_button.set_sensitive(False)
            self.previous_file_button_revealer.set_reveal_child(False)
            self.next_file_button.set_sensitive(False)
            self.next_file_button_revealer.set_reveal_child(False)
        else:
            if self.file_manager.selected_files.get_n_items() > 1:
                self.previous_file_button.set_sensitive(True)
                self.next_file_button.set_sensitive(True)
            else:
                self.previous_file_button.set_sensitive(False)
                self.next_file_button.set_sensitive(False)
            self.previous_file_button_revealer.set_reveal_child(True)
            self.next_file_button_revealer.set_reveal_child(True)

    def update_binds(self, *args):
        """
        Reads the file data from the file manager and applies it
        to the file view.
        """
        self.update_buttons()
        asyncio.create_task(self._update_binds())

    bind_in_progress = GObject.Property(type=bool, default=False)

    async def _update_binds(self):
        self.bind_in_progress = True

        _selected_files = set(self.file_manager.selected_files_list)

        # Get list of selected (added)/unselected (removed) files
        added_files = _selected_files - self.files
        removed_files = self.files - _selected_files

        # Handle added and removed files
        await self._unbind_files(removed_files)
        await self._bind_files(added_files)

        # We show/hide new entries only as necessary once the bind/unbind are
        # done; otherwise the entry is hidden, cleared, then shown and uncleared,
        # which causes a distracting animation.
        self.more_tags_group.refresh_entry_visibility()

        # Make save/fields sensitive/insensitive based on whether selected files are
        # all writable
        has_unwritable = False
        for file in _selected_files:
            if not file.is_writable:
                has_unwritable = True
                break

        if has_unwritable:
            self.album_cover.set_sensitive(False)
            self.important_data_container.set_sensitive(False)
            self.tag_list.set_sensitive(False)
        else:
            self.album_cover.set_sensitive(True)
            self.important_data_container.set_sensitive(True)
            self.tag_list.set_sensitive(True)

        # Scroll to the top of the view
        adjust = self.content_scroll.get_vadjustment()
        adjust.set_value(adjust.get_lower())

        self.bind_in_progress = False

    async def _bind_files(self, files):
        """Binds a file to the fileview. Used internally in update_binds."""
        if not files:
            return

        self.files |= set(files)
        await self.tagentry_manager.bind_to_files(files)

        for entry in self.bindable_entries:
            entry.bind_to_files(files)

        await self.more_tags_group.bind_to_files(files)

        self.file_info.refresh_label()

    async def _unbind_files(self, files):
        """Unbinds a file from the fileview. Used internally in update_binds."""
        if not files:
            return

        self.files -= set(files)
        await self.tagentry_manager.unbind_from_files(files)

        for entry in self.bindable_entries:
            entry.unbind_from_files(files)

        await self.more_tags_group.unbind_from_files(files)

        self.file_info.refresh_label()

    @GObject.Property(type=bool, default=False)
    def compact(self):
        return self.more_tags_group.compact

    @compact.setter
    def compact(self, value):
        self.more_tags_group.compact = value

    @GObject.Property(type=bool, default=False)
    def height_below_360(self):
        """Hack to fix https://gitlab.gnome.org/World/eartag/-/merge_requests/130"""
        return self.more_tags_group.height_below_360

    @height_below_360.setter
    def height_below_360(self, value):
        self.more_tags_group.height_below_360 = value



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from ..backends.file import VALID_TAGS

from gi.repository import Adw, Pango, GObject
import re

THEMES = {
    "light": {
        "bracket_color": (61, 56, 70),
        "placeholder_colors": (
            (26, 95, 180),  # @blue_5
            (38, 162, 105),  # @green_5
            (229, 165, 10),  # @yellow_5
            (198, 70, 0),  # @orange_5
            (165, 29, 45),  # @red_5
            (97, 53, 131),  # @purple_5
            (99, 69, 44),  # @brown_5
        ),
    },
    "dark": {
        "bracket_color": (154, 153, 150),
        "placeholder_colors": (
            (153, 193, 241),  # @blue_1
            (143, 240, 164),  # @green_1
            (249, 240, 107),  # @yellow_1
            (255, 190, 111),  # @orange_1
            (246, 97, 81),  # @red_1
            (220, 138, 221),  # @purple_1
            (181, 131, 90),  # @brown_2
        ),
    },
}


def pango_attr_iter(attrlist: Pango.AttrList):
    """Provides an iterator for Pango.AttrList."""
    iter = attrlist.get_iterator()
    try:
        attr = iter.get_attrs()[0]
    except IndexError:
        return
    while attr:
        yield attr
        iter.next()
        try:
            attr = iter.get_attrs()[0]
        except IndexError:
            break


def attr_foreground_new(color: (int, int, int), start_index, end_index) -> Pango.Attribute:
    """Shorthand function to create a new Pango AttrForeground."""
    attr = Pango.attr_foreground_new((color[0]) * 256, (color[1]) * 256, (color[2]) * 256)
    attr.start_index = start_index
    attr.end_index = end_index
    return attr


def attr_underline_new(start_index, end_index) -> Pango.Attribute:
    """Shorthand function to create a new Pango AttrUnderline."""
    attr = Pango.attr_underline_new(Pango.Underline.SINGLE)
    attr.start_index = start_index
    attr.end_index = end_index
    return attr


class EartagPlaceholderSyntaxHighlighter(GObject.Object):
    """Helper object for placeholder syntax highlighting."""

    error = GObject.Property(type=bool, default=False)

    theme = GObject.Property(type=str, default="light")
    high_contrast = GObject.Property(type=bool, default=False)

    def __init__(self, widget, widget_type="entry", allow_duplicates=False):
        """
        Sets up syntax highlighting for a widget.
        """
        super().__init__()

        self._tag_positions_changed = False
        self._tag_positions = []

        self.widget = widget
        self.widget_type = widget_type
        self.allow_duplicates = allow_duplicates

        if widget_type == "entry":
            widget.get_delegate().get_buffer().connect(
                "inserted-text", self.syntax_highlighting_inserted
            )
            widget.get_delegate().get_buffer().connect(
                "deleted-text", self.syntax_highlighting_deleted
            )
        elif widget_type == "label":
            widget.connect("notify::label", self.syntax_highlighting_label_updated)

        self.widget_type = "entry"

        style_manager = Adw.StyleManager.get_default()
        style_manager.connect("notify::dark", self.update_theme)
        style_manager.connect("notify::high-contrast", self.update_theme)
        self.update_theme(style_manager)

        self.update_syntax_highlighting()

    def get_text(self):
        if self.widget_type == "entry":
            return self.widget.get_text()
        elif self.widget_type == "label":
            return self.widget.get_label()

    def update_theme(self, style_manager, *args):
        self.props.theme = "dark" if style_manager.props.dark else "light"
        self.props.high_contrast = style_manager.props.high_contrast
        self.update_syntax_highlighting()

    def update_syntax_highlighting(self, full_text=None):
        """Helper function that returns tag positions."""
        attrs = Pango.AttrList()
        error = False

        # Pango attributes (used for syntax highlighting) use offsets calculated
        # in bytes, not Python characters, so we encode the pattern and filename
        # to UTF-8 so that the returned group positions match the byte count.
        if full_text is None:
            full_text = self.get_text()
        full_text = full_text.encode("utf-8")

        n = 0
        present_tags = set()
        for match in re.finditer(r"{.*?}".encode("utf-8"), full_text):
            try:
                tag_name = match.group(0)[1:-1].decode("utf-8")
            except (IndexError, UnicodeDecodeError):
                error = True
                continue

            if "{" in tag_name or "}" in tag_name:
                error = True
                continue

            if tag_name == "" or (not self.allow_duplicates and tag_name in present_tags):
                continue

            if tag_name not in VALID_TAGS + ("length", "bitrate"):
                continue

            if not self.allow_duplicates:
                present_tags.add(tag_name)

            # Add bracket colors
            for position in (match.span(0)[0], match.span(0)[1] - 1):
                color_attr = attr_foreground_new(
                    THEMES[self.theme]["bracket_color"], position, position + 1
                )
                attrs.insert(color_attr)

            # Add tag color
            color = THEMES[self.theme]["placeholder_colors"][
                n % len(THEMES[self.theme]["placeholder_colors"])
            ]
            color_attr = attr_foreground_new(color, match.span(0)[0] + 1, match.span(0)[1] - 1)
            attrs.insert(color_attr)

            n += 1

        self.props.error = error

        self.widget.set_attributes(attrs)

    def syntax_highlighting_inserted(self, *args):
        """
        Updates the placeholder syntax highlighting after text insertion.
        """
        self.update_syntax_highlighting()

    def syntax_highlighting_deleted(self, entry, position: int, n_chars: int):
        """
        Updates the placeholder syntax highlighting after text removal.
        """
        full_text = entry.get_text()
        full_text = full_text[:position] + full_text[position + n_chars :]
        self.update_syntax_highlighting(full_text)

    def syntax_highlighting_label_updated(self, label, *args):
        self.update_syntax_highlighting(label.get_label())



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

import sys
import gi
import os.path
import logging

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("GdkPixbuf", "2.0")

from gi.repository import Adw, Gtk, Gio

# asyncio setup
import asyncio
from ._async import policy

from . import APP_ID, APP_GRESOURCE_PATH
from .utils.validation import is_valid_music_file, get_mimetype
from .window import EartagFileDialogType, EartagWindow
from .logger import logger


class Application(Adw.Application):
    def __init__(self, version="dev", devel=False):
        super().__init__(
            application_id=APP_ID,
            resource_base_path=APP_GRESOURCE_PATH,
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
        )
        self.version = version
        self.devel = devel
        self.paths = []
        self.connect("open", self.on_open)
        if devel:
            logger.setLevel(logging.DEBUG)
        logger.info(f"Starting Ear Tag {version}{' (development mode)' if devel else ''}")

    def on_open(self, app, files, *args):
        window = self.props.active_window
        if window:
            paths = []
            for file in files:
                path = file.get_path()
                if path:
                    paths.append(path)
            window.file_manager.load_files(paths)
        else:
            for file in files:
                path = file.get_path()
                if path:
                    if not os.path.exists(path):
                        continue
                    if os.path.isdir(path):
                        for _file in os.listdir(path):
                            _fpath = os.path.join(path, _file)
                            if os.path.isfile(_fpath) and is_valid_music_file(_fpath):
                                self.paths.append(_fpath)
                        continue
                    elif not is_valid_music_file(path):
                        continue
                    self.paths.append(path)
        self.do_activate()

    def do_activate(self):
        win = self.props.active_window
        if not win:
            win = EartagWindow(application=self, paths=self.paths, devel=self.devel)
            self.create_action("preferences", self.on_preferences_action, None)
            self.create_action("about", self.on_about_action, None)

            self.create_action("open_file", self.on_open_file_action, "<Ctrl>o")
            self.create_action("save", self.on_save_action, "<Ctrl>s")
            self.create_action("open_folder", self.on_open_folder_action, "<Ctrl>d")

            self.create_action("next_file", self.on_next_action, "<Alt>Right")
            self.create_action("previous_file", self.on_previous_action, "<Alt>Left")
            self.create_action("close_selected", self.on_close_selected_action, "<Ctrl>w")
            self.select_all_action = self.create_action(
                "select_all", self.on_select_all_action, "<Ctrl><Shift>a"
            )

            self.create_action("toggle_sidebar", self.on_toggle_sidebar_action, "F9")
            self.create_action("open_menu", self.on_open_menu_action, "F10")

            self.rename_action = self.create_action("rename", self.on_rename_action, None)
            self.rename_action.set_enabled(False)

            self.extract_action = self.create_action("extract", self.on_extract_action, None)
            self.extract_action.set_enabled(False)

            self.identify_action = self.create_action("identify", self.on_identify_action, None)
            self.identify_action.set_enabled(False)

            self.undo_all_action = self.create_action("undo_all", self.on_undo_all_action, None)
            self.undo_all_action.set_enabled(False)

            self.delete_all_tags_action = self.create_action(
                "delete_all_tags", self.on_delete_all_tags_action, None
            )
            self.delete_all_tags_action.set_enabled(False)

            self.create_action("quit", self.on_quit_action, "<Ctrl>q")

            self._ = _

        win.present()

    def create_action(self, name, callback, accel=None):
        """Add an Action and connect to a callback"""
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)
        if accel:
            self.set_accels_for_action(f"app.{name}", (accel, None))
        return action

    def on_save_action(self, widget, _):
        self.get_active_window().file_manager.save()

    def on_rename_action(self, widget, _):
        self.get_active_window().show_rename_dialog()

    def on_identify_action(self, widget, _):
        self.get_active_window().show_identify_dialog()

    def on_preferences_action(self, widget, _):
        self.get_active_window().show_preferences_dialog()

    def on_extract_action(self, widget, _):
        self.get_active_window().show_extract_dialog()

    def on_undo_all_action(self, widget, _):
        self.get_active_window().undo_all()

    def on_delete_all_tags_action(self, widget, _):
        self.get_active_window().show_delete_all_tags_dialog()

    def on_about_action(self, widget, _):
        version_str = self.version
        if self.devel:
            version_str += " (dev)"

        about = Adw.AboutDialog(
            application_name=self._("Ear Tag"),
            application_icon=APP_ID,
            developers=["knuxify https://tilde.town/~knuxify"],
            artists=["Jakub Steiner", "Igor Dyatlov"],
            license_type=Gtk.License.MIT_X11,
            issue_url="https://gitlab.gnome.org/World/eartag/-/issues",
            version=version_str,
            website="https://gitlab.gnome.org/World/eartag",
        )

        if self._("translator-credits") != "translator-credits":
            # TRANSLATORS: Add your name/nickname here
            about.props.translator_credits = self._("translator-credits")

        lib_versions = []

        lib_versions.append(
            f"gtk4: {Gtk.get_major_version()}.{Gtk.get_minor_version()}.{Gtk.get_micro_version()}"  # noqa: E501
        )
        lib_versions.append(
            f"libadwaita: {Adw.get_major_version()}.{Adw.get_minor_version()}.{Adw.get_micro_version()}"  # noqa: E501
        )

        import filetype

        lib_versions.append(f"filetype.py: {filetype.version}")

        import mutagen

        lib_versions.append(f"mutagen: {mutagen.version_string}")

        import PIL

        lib_versions.append(f"pillow: {PIL.__version__}")

        import aiohttp

        lib_versions.append(f"aiohttp: {aiohttp.__version__}")

        import xxhash

        lib_versions.append(
            f"xxhash: {xxhash.VERSION} (using library version {xxhash.XXHASH_VERSION})"
        )

        lib_version_str = "\n - ".join(lib_versions)

        opened_file_list = []
        for file in self.props.active_window.file_manager.files:
            opened_file_list.append(
                f"{file.path}, {get_mimetype(file.path)}, {file.__gtype_name__}"
            )  # noqa: E501

        opened_file_list_str = "\n - ".join(opened_file_list) or "None"

        about.set_debug_info(
            f"""Ear Tag {self.version}{" (Development version)" if self.devel else ""}

Running in Flatpak: {os.path.exists("/.flatpak-info") and "YES" or "NO"}

Dependency versions:
 - {lib_version_str}

Opened files:
 - {opened_file_list_str}"""
        )

        about.present(self.props.active_window)

    def on_open_file_action(self, widget, _):
        window = self.get_active_window()
        window.show_file_chooser(
            EartagFileDialogType.TYPE_FILE | EartagFileDialogType.MODE_OVERWRITE
        )

    def on_open_folder_action(self, widget, _):
        window = self.get_active_window()
        window.show_file_chooser(
            EartagFileDialogType.TYPE_FOLDER | EartagFileDialogType.MODE_OVERWRITE
        )

    def on_next_action(self, *args):
        win = self.props.active_window
        win.select_next()

    def on_previous_action(self, *args):
        win = self.props.active_window
        win.select_previous()

    def on_close_selected_action(self, *args):
        win = self.props.active_window
        if win.file_manager.files:
            win.remove_selected()
        else:
            win.close()

    def on_select_all_action(self, *args):
        win = self.props.active_window
        if win.container_stack.get_visible_child() != win.split_view:
            return
        if win.file_manager.all_selected():
            if win.file_manager.files.get_n_items() == 1:
                if not win.select_multiple_button.get_active():
                    win.select_multiple_button.set_active(True)
                    return
            win.select_multiple_button.set_active(False)
            win.file_manager.unselect_all()
            win.file_manager.emit("select-first")
        else:
            win.select_multiple_button.set_active(True)
            win.file_manager.select_all()

    def on_toggle_sidebar_action(self, *args):
        win = self.props.active_window
        if win.split_view.get_collapsed():
            if win.split_view.get_show_sidebar():
                win.hide_sidebar()
            else:
                win.show_sidebar()
        else:
            win.sidebar_file_list.grab_focus()

    def on_open_menu_action(self, *args):
        win = self.props.active_window
        if win.container_stack.get_visible_child() == win.split_view:
            win.primary_menu_button.activate()
        else:
            win.empty_primary_menu_button.activate()

    def on_quit_action(self, *args):
        win = self.props.active_window
        win.close()


def main(version, devel):
    asyncio.set_event_loop_policy(policy)
    app = Application(version, devel)
    return app.run(sys.argv)
